/****************************************************************************
 * OBIDMS_column_seq functions                                              *
 ****************************************************************************/

/**
 * @file obidsmcolumn_seq.c
 * @author Celine Mercier
 * @date November 18th 2015
 * @brief Functions handling OBIColumns containing data in the form of indices referring to DNA sequences.
 */


#include <stdlib.h>
#include <stdio.h>

#include "obidmscolumn.h"
#include "obiview.h"
#include "obitypes.h"
#include "obierrno.h"
#include "obidebug.h"
#include "obiavl.h"


#define DEBUG_LEVEL 0	// TODO has to be defined somewhere else (cython compil flag?)


/**********************************************************************
 *
 * D E F I N I T I O N   O F   T H E   P U B L I C   F U N C T I O N S
 *
 **********************************************************************/


int obi_column_set_obiseq_with_elt_idx(OBIDMS_column_p column, index_t line_nb, index_t element_idx, char* value)
{
	byte_t* value_b;
	index_t idx;

	// Check that the line number is not greater than the maximum allowed
	if (line_nb >= MAXIMUM_LINE_COUNT)
	{
		obi_set_errno(OBICOL_UNKNOWN_ERROR);
		obidebug(1, "\nError trying to set a value at a line number greater than the maximum allowed");
    	return -1;
	}

	// Check if the file needs to be enlarged
	while ((line_nb+1) > (column->header)->line_count)
	{
		// Enlarge the file
		if (obi_enlarge_column(column) < 0)
	    	return -1;
	}

	// Update lines used
	if ((line_nb+1) > (column->header)->lines_used)
		(column->header)->lines_used = line_nb+1;

	// Encode the value on a byte array with a header
	value_b = obi_seq_to_obibytes(value);
	if (value_b == NULL)
		return -1;

	if (strlen(value_b) == 0)
		fprintf(stderr, "\nPOUIC");

	//fprintf(stderr, "\n>%s||%s", value, obi_obibytes_to_seq(value_b));

	// Add in the AVL tree
	idx = obi_avl_add((column->avl)[compute_crc(value)], value_b);
	if (idx == -1)
		return -1;

	// Add the value's index in the column
	*(((index_t*) (column->data)) + (line_nb * ((column->header)->nb_elements_per_line)) + element_idx) = idx;

	free(value_b);

	return 0;
}


int obi_column_set_obiseq_with_elt_idx_in_view(Obiview_p view, OBIDMS_column_p column, index_t line_nb, index_t element_idx, char* value)
{
	// Check that the view is not read-only
	if (view->read_only)
	{
		obi_set_errno(OBIVIEW_ERROR);
		obidebug(1, "\nError trying to set a value in a column in a read-only view");
		return -1;
	}

	if ((view->line_selection != NULL) || (!(column->writable)))
	{
		// Get the right line number
		if (column->writable)
			line_nb = *(((index_t*) ((view->line_selection)->data)) + line_nb);

		column = obi_view_clone_column(view, (column->header)->name);
		if (column == NULL)
		{
			obidebug(1, "\nError trying to clone a column to modify it");
			return -1;
		}
	}

	if ((line_nb+1) > view->line_count)
	{
		if (obi_view_update_lines(view, (line_nb+1)) < 0)
			return -1;
	}

	if (obi_column_set_obiseq_with_elt_idx(column, line_nb, element_idx, value) < 0)
		return -1;

	return 0;
}


const char* obi_column_get_obiseq_with_elt_idx(OBIDMS_column_p column, index_t line_nb, index_t element_idx)
{
	index_t idx;
	byte_t* value_b;

	if ((line_nb+1) > ((column->header)->line_count))
	{
		obi_set_errno(OBICOL_UNKNOWN_ERROR);
		obidebug(1, "\nError trying to get a value that is beyond the current number of lines used in the column");
		return OBISeq_NA;
	}

	idx = *(((index_t*) (column->data)) + (line_nb * ((column->header)->nb_elements_per_line)) + element_idx);

	// Check NA
	if (idx == OBIIdx_NA)
		return OBISeq_NA;

	//value_b = obi_avl_get((column->avl)[crc(value)], idx);
	return obi_obibytes_to_seq(value_b);
}


const char* obi_column_get_obiseq_with_elt_idx_in_view(Obiview_p view, OBIDMS_column_p column, index_t line_nb, index_t element_idx)
{
	if ((line_nb+1) > (view->line_count))
	{
		obi_set_errno(OBICOL_UNKNOWN_ERROR);
		obidebug(1, "\nError trying to get a value that is beyond the current line count of the view");
		return OBISeq_NA;
	}

	if (view->line_selection)
		line_nb = *(((index_t*) ((view->line_selection)->data)) + line_nb);

	return obi_column_get_obiseq_with_elt_idx(column, line_nb, element_idx);
}


int obi_column_set_obiseq_with_elt_name(OBIDMS_column_p column, index_t line_nb, const char* element_name, char* value)
{
	index_t element_idx;
	element_idx = obi_column_get_element_index_from_name(column, element_name);
	if (element_idx == OBIIdx_NA)
		return -1;
	obi_column_set_obiseq_with_elt_idx(column, line_nb, element_idx, value);
	return 0;
}


int obi_column_set_obiseq_with_elt_name_in_view(Obiview_p view, OBIDMS_column_p column, index_t line_nb, const char* element_name, char* value)
{
	index_t element_idx;
	element_idx = obi_column_get_element_index_from_name(column, element_name);
	if (element_idx == OBIIdx_NA)
		return -1;
	obi_column_set_obiseq_with_elt_idx_in_view(view, column, line_nb, element_idx, value);
	return 0;
}


const char* obi_column_get_obiseq_with_elt_name(OBIDMS_column_p column, index_t line_nb, const char* element_name)
{
	index_t element_idx;

	element_idx = obi_column_get_element_index_from_name(column, element_name);
	if (element_idx == OBIIdx_NA)
		return OBISeq_NA;
	return obi_column_get_obiseq_with_elt_idx(column, line_nb, element_idx);
}


const char* obi_column_get_obiseq_with_elt_name_in_view(Obiview_p view, OBIDMS_column_p column, index_t line_nb, const char* element_name)
{
	index_t element_idx;

	element_idx = obi_column_get_element_index_from_name(column, element_name);
	if (element_idx == OBIIdx_NA)
		return OBISeq_NA;
	return obi_column_get_obiseq_with_elt_idx_in_view(view, column, line_nb, element_idx);
}

